/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/

package casa.dodwan.napi.ws;

import casa.dodwan.napi.CommandExecutor;
import casa.dodwan.napi.CommandExecutor.CommandException;
import casa.dodwan.run.Dodwan;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import casa.dodwan.napi.DodwanNapiEndpoint;
import casa.dodwan.napi.DodwanPdu;
import casa.util.pdu.PduSerializer;
import java.io.ByteArrayInputStream;
import java.nio.Buffer;
import java.nio.ByteBuffer;
import org.java_websocket.WebSocket;
import org.java_websocket.exceptions.WebsocketNotConnectedException;

/**
 * Websocket endpoint that implements the DoDWAN Network API endpoint
 */

public class DodwanWsEndpoint implements DodwanNapiEndpoint {
      
    public Dodwan dodwan ;
    private String clientAddr;
    private CommandExecutor commandExecutor;
    private InputStream inStream;
    private PduSerializer serializer;
    private WebSocket conn;
    private String name;
    
    
    public DodwanWsEndpoint(WebSocket conn, String endpointName) {
        this.conn = conn;
        this.name = endpointName;
        this.clientAddr = conn.getRemoteSocketAddress().getAddress().getHostAddress();
        
        this.dodwan = DodwanWsServer.dodwan;
        this.serializer = DodwanWsServer.pduSerializer;
        
        this.commandExecutor = new CommandExecutor(this);
        //this.commandExecutor.init();
        
    }
    
    
    
    //-------------------------------------------------------------------------
    /**
     *  Closes the connection this endpoint is using     */
    @Override
    public void closeConnection() {
        this.conn.close();
    }
    
    
    //-------------------------------------------------------------------------
    /**
     *  Returns the address of the client
     */
    public String getClientAddr() {
        return this.clientAddr;
    }
    
    
    //-------------------------------------------------------------------------
    /**
     *  Returns the name of the endpoint
     */
    public String getName() {
        return this.name;
    }
    
    
     //-------------------------------------------------------------------------
    /**
     *  Does the necessary work before the endpoint can be destroyed
     */
    //public void end() {
    //    this.commandExecutor.end();
    //}
   
    //-------------------------------------------------------------------------
    /**
     *  Handles the reception of a message
     *  
     */
    public void rcvMsg(ByteBuffer message) {
        //System.err.println("DodwanWsEndpoint: received message:  "+new String(message.array()));
        
        try {
           this.inStream = new ByteArrayInputStream(message.array());
        }
        catch (Exception e) {
            System.err.println("DodwanWsEndpoint: IO error when transforming received buffer into ByteArrayInputStream :");
            e.printStackTrace(System.err);
        }
        
        try {
            this.commandExecutor.receiveAndExecuteCommand();    
        }
        catch (CommandException ce) {
               System.err.println("DodwanWsEndpoint command error when receiving and executing command: "+ce.getMessage());
        }
        catch (IOException ioe) {
            //  If the message is not understandable (wrong serialialization), an error PDU without token is sent back.
                System.err.println("DodwanWsEndpoint IO or parsing error when receiving command");
                try {
                    this.commandExecutor.sendError(null,"IO or parsing error on PDU");
                }
                catch (IOException e) {
                   System.err.println("DodwanWsEndpoint: tried to sendError, but failed"); 
                   e.printStackTrace(System.err);
                }
        } 
        
    }
    
    
   
    
      
    //-------------------------------------------------------------------------
    /*
    * Receive a PDU from the client
    */
    @Override
    public  DodwanPdu receivePdu()  throws IOException {
        DodwanPdu pdu = new DodwanPdu();
        if (serializer.read(pdu, this.inStream) == -1 ) {
           throw new IOException(); 
        }
        System.err.println("DodwanEndpoint("+this.name+") receiving PDU : "+pdu);
        return pdu;
    }
    
      
    //-------------------------------------------------------------------------
    /**
     * Sends a PDU to the client
     * A single buffer is allocated for the entire PDU (assumed to be of reasonable size)
     */
    @Override
    public  void sendPdu(DodwanPdu pdu) {
        System.err.println("DodwanEndpoint("+this.name+") sending PDU : "+pdu);
        int bsize = serializer.getSerializedSize(pdu);
        //System.err.println("DodwanWsEndpoint allocate outBuffer with size "+bsize);
        ByteBuffer outBuffer = ByteBuffer.allocate(bsize);
        
        //System.err.println("DodwanWsEndpoint sending PDU " + pdu.getString("name"));
        ByteBufferOutputStream outStream = new ByteBufferOutputStream(outBuffer);
        
        try {
            serializer.write(pdu, outStream);
            outStream.flush();
            //System.err.println("DodwanWsEndpoint filled outBuffer "+outBuffer.toString());

        } catch (IOException e) {
            System.err.println("Error when filling buffer for PDU " + pdu.getString("name"));
            e.printStackTrace(System.err);
        }

        try { 
            ((Buffer) outBuffer).flip();  // cast to Buffer added to make compabible with java 8 and java > 9
            //System.err.println("DodwanWsEndpoint sending outBuffer "+outBuffer.toString()+ " : "+new String(outBuffer.array()));          
            this.conn.send(outBuffer);        
        } catch (WebsocketNotConnectedException wnce) {
            System.err.println("Connection closed. Can't send PDU " + pdu.getString("name"));
            //wnce.printStackTrace(System.err);
        } catch (IllegalArgumentException iae) {
            System.err.println("Null buffer error when trying to send PDU " + pdu.getString("name"));
            iae.printStackTrace(System.err);
        }
    }


    //-------------------------------------------------------------------------
    @Override
    public Dodwan getDodwan() {
        return this.dodwan;
    }

    
    
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////
    
    
    /**
     * Allows us to reuse an existing ByteBuffer while writing to an output stream
     */
    private class ByteBufferOutputStream extends OutputStream {

        private ByteBuffer byteBuffer;

        // Creates an uninitialized stream (must use setByteBuffer afterwards) 
        public ByteBufferOutputStream() {
        }

        // Creates a stream with a new non-direct buffer 
        public ByteBufferOutputStream(int bufferSize) {
            this(ByteBuffer.allocate(bufferSize));
        }

        // Creates a stream backed by a given buffer 
        public ByteBufferOutputStream(ByteBuffer byteBuffer) {
            this.byteBuffer = byteBuffer;
        }

        public ByteBuffer getByteBuffer() {
            return byteBuffer;
        }

        public void setByteBuffer(ByteBuffer byteBuffer) {
            this.byteBuffer = byteBuffer;
        }

        
        
        public void write(int b) throws IOException {
            if (!byteBuffer.hasRemaining()) {
                flush();
            }
            byteBuffer.put((byte) b);
        }

        public void write(byte[] bytes, int offset, int length) throws IOException {
            //System.err.print("ByteBufferOutputStream writing byte[] offset="+ offset + " length="+ length+ "   ");
            //for (int k=offset; k<(offset+length-1); k++) { System.err.print((char)(bytes[k])); } ; System.err.println();
            if (byteBuffer.remaining() < length) {
                flush();
            }
            byteBuffer.put(bytes, offset, length);
            //System.err.println("ByteBufferOutputStream written : "+byteBuffer);
        }
    }


    
}

